import os
import json
import gzip
import math

from io import TextIOWrapper

from datetime import datetime, time

from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db import router, transaction
from django.utils.timezone import utc
from django.utils.timezone import now
from django.utils.module_loading import import_string

from celery import current_app
from celery.exceptions import NotRegistered

from chamber.shortcuts import change_and_save

from security.backends.writer import BaseBackendWriter
from security.config import settings
from security.enums import RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState

from .models import CeleryTaskInvocationLog, get_log_model_from_logger_name, get_response_state


MAX_VERSION = 9999


def lazy_serialize_qs_without_pk(qs):
    def generator():
        for obj in qs.iterator():
            data = serializers.serialize('python', [obj])[0]
            yield data

    class StreamList(list):
        def __iter__(self):
            return generator()

        def __len__(self):
            return 1

    return StreamList()


def get_querysets_by_batch(qs, batch):
    steps = math.ceil(qs.count() / batch)
    for _ in range(steps):
        yield qs[:batch]


class SQLBackendWriter(BaseBackendWriter):

    def _set_related_objects(self, log, logger):
        log.related_objects.add(
            *[related_object_triple[1:] for related_object_triple in logger.related_objects]
        )

    def _get_version(self, logger, is_last):
        if is_last:
            return MAX_VERSION
        elif 'last_sql_version' in logger.backend_logs:
            return logger.backend_logs['last_sql_version'] + 1
        else:
            return 0

    def _create_from_logger(self, logger, is_last, **extra_data):
        version = self._get_version(logger, is_last)
        data = logger.to_dict()
        del data['related_objects']

        model_class = get_log_model_from_logger_name(logger.logger_name)
        if not model_class.objects.filter(id=logger.id, version__gte=version).exists():
            instance = model_class(
                version=version,
                **extra_data,
                **data
            )
            instance.save()
            self._set_related_objects(instance, logger)
            logger.backend_logs['sql_instance'] = instance
            logger.backend_logs['last_sql_version'] = version

    def _update_from_logger(self, logger, is_last, **extra_data):
        version = self._get_version(logger, is_last)
        data = logger.to_dict()
        del data['related_objects']

        with transaction.atomic(using=settings.LOG_DB_NAME):
            instance = logger.backend_logs['sql_instance'].__class__.objects.select_for_update().get(id=logger.id)
            if instance.version < version:
                instance = change_and_save(
                    logger.backend_logs['sql_instance'],
                    version=version,
                    **data,
                    **extra_data,
                    update_only_changed_fields=True,
                )
                self._set_related_objects(instance, logger)
            logger.backend_logs['sql_instance'] = instance
            logger.backend_logs['last_sql_version'] = version

    def _create_or_update_from_logger(self, logger, is_last=False, **extra_data):
        if 'sql_instance' in logger.backend_logs:
            self._update_from_logger(logger, is_last, **extra_data)
        else:
            self._create_from_logger(logger, is_last, **extra_data)

    def input_request_started(self, logger):
        self._create_or_update_from_logger(logger, state=RequestLogState.INCOMPLETE)

    def input_request_finished(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=get_response_state(logger.response_code)
        )

    def input_request_error(self, logger):
        self._create_or_update_from_logger(logger, state=RequestLogState.ERROR)

    def output_request_started(self, logger):
        self._create_or_update_from_logger(logger, state=RequestLogState.INCOMPLETE)

    def output_request_finished(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=get_response_state(logger.response_code)
        )

    def output_request_error(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=RequestLogState.ERROR
        )

    def command_started(self, logger):
        self._create_or_update_from_logger(
            logger, state=CommandState.ACTIVE
        )

    def command_output_updated(self, logger):
        self._create_or_update_from_logger(
            logger, state=CommandState.ACTIVE
        )

    def command_finished(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CommandState.SUCCEEDED
        )

    def command_error(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CommandState.FAILED
        )

    def celery_task_invocation_started(self, logger):
        self._create_or_update_from_logger(
            logger, state=CeleryTaskInvocationLogState.WAITING
        )

    def celery_task_invocation_triggered(self, logger):
        self._create_or_update_from_logger(
            logger, state=CeleryTaskInvocationLogState.TRIGGERED
        )

    def celery_task_invocation_duplicate(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.DUPLICATE
        )

    def celery_task_invocation_ignored(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.IGNORED
        )

    def celery_task_invocation_timeout(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.TIMEOUT
        )

    def celery_task_invocation_expired(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.EXPIRED
        )

    def celery_task_invocation_succeeded(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.SUCCEEDED
        )

    def celery_task_invocation_failed(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.FAILED
        )

    def celery_task_run_started(self, logger):
        self._create_or_update_from_logger(
            logger, state=CeleryTaskRunLogState.ACTIVE
        )

    def celery_task_run_succeeded(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskRunLogState.SUCCEEDED
        )

    def celery_task_run_failed(self, logger):
        self._create_or_update_from_logger(
            logger, is_last=True, state=CeleryTaskRunLogState.FAILED
        )

    def celery_task_run_retried(self, logger):
        self._create_or_update_from_logger(
            logger, state=CeleryTaskRunLogState.RETRIED
        )

    def celery_task_run_output_updated(self, logger):
        self._create_or_update_from_logger(
            logger, state=CeleryTaskRunLogState.ACTIVE
        )

    def set_stale_celery_task_log_state(self):
        processsing_stale_tasks = CeleryTaskInvocationLog.objects.filter_processing(
            stale_at__lt=now()
        ).order_by('stale_at')
        for task in processsing_stale_tasks[:settings.SET_STALE_CELERY_INVOCATIONS_LIMIT_PER_RUN]:
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                change_and_save(
                    task,
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=task_last_run.stop,
                    time=(task_last_run.stop - task.start).total_seconds(),
                    version=MAX_VERSION,
                    update_only_changed_fields=True,
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                change_and_save(
                    task,
                    state=CeleryTaskInvocationLogState.FAILED,
                    stop=task_last_run.stop,
                    time=(task_last_run.stop - task.start).total_seconds(),
                    version=MAX_VERSION,
                    update_only_changed_fields=True
                )
            else:
                try:
                    current_app.tasks[task.name].expire_invocation(
                        task.id,
                        task.task_args,
                        task.task_kwargs,
                        dict(
                            slug=task.slug,
                            parent_log=task.parent_log,
                            related_objects=[
                                (
                                    router.db_for_write(related_object.object_ct.model_class()),
                                    related_object.object_ct_id,
                                    related_object.object_id
                                ) for related_object in task.related_objects.all()
                            ],
                            start=task.start,
                            celery_task_id=task.celery_task_id,
                            stop=task.stop,
                            name=task.name,
                            queue_name=task.queue_name,
                            applied_at=task.applied_at,
                            triggered_at=task.triggered_at,
                            is_unique=task.is_unique,
                            is_async=task.is_async,
                            is_on_commit=task.is_on_commit,
                            input=task.input,
                            task_args=task.task_args,
                            task_kwargs=task.task_kwargs,
                            estimated_time_of_first_arrival=task.estimated_time_of_first_arrival,
                            expires_at=task.expires_at,
                            stale_at=task.stale_at,
                        )
                    )
                except NotRegistered:
                    change_and_save(
                        task,
                        state=CeleryTaskInvocationLogState.EXPIRED,
                        stop=now(),
                        version=MAX_VERSION,
                        update_only_changed_fields=True
                    )

    def clean_logs(self, type, timestamp, backup_path, stdout, stderr):
        storage = import_string(settings.BACKUP_STORAGE_CLASS)()

        qs = get_log_model_from_logger_name(type).objects.filter(stop__lte=timestamp).order_by('stop')
        for step_timestamp in qs.datetimes('start', 'day', tzinfo=utc):
            min_timestamp = datetime.combine(step_timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(step_timestamp, time.max).replace(tzinfo=utc)
            qs_filtered_by_day = qs.filter(stop__range=(min_timestamp, max_timestamp))

            for qs_batch in get_querysets_by_batch(qs_filtered_by_day, settings.CLEAN_LOGS_BACKUP_FILE_BATCH_SIZE):
                stdout.write(
                    2 * ' ' + 'Cleaning logs for date {} ({})'.format(
                        step_timestamp.date(), qs_batch.count()
                    )
                )
                if backup_path:
                    log_file_name = os.path.join(backup_path, str(step_timestamp.date()))

                    if storage.exists('{}.json.gz'.format(log_file_name)):
                        i = 1
                        while storage.exists('{}({}).json.gz'.format(log_file_name, i)):
                            i += 1
                        log_file_name = f'{log_file_name}({i})'

                    stdout.write(4 * ' ' + 'generating backup file: {}.json.gz'.format(log_file_name))

                    with storage.open('{}.json.gz'.format(log_file_name), 'wb') as f:
                        with TextIOWrapper(gzip.GzipFile(filename='{}.json'.format(log_file_name),
                                                         fileobj=f, mode='wb')) as gzf:
                            json.dump(
                                lazy_serialize_qs_without_pk(qs_batch), gzf, cls=DjangoJSONEncoder, indent=5
                            )

                stdout.write(4 * ' ' + 'deleting logs')
                for qs_batch_to_delete in get_querysets_by_batch(qs_batch, settings.SQL_CLEAN_LOGS_DELETE_BATCH_SIZE):
                    qs_filtered_by_day.filter(pk__in=qs_batch_to_delete).delete()



import os

from setuptools import setup, find_packages

from security.version import get_version


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='django-security-logger',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    version=get_version(),
    description="Django security library.",
    keywords='django, throttling',
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/django-security',
    license='MIT',
    package_dir={'security': 'security'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Czech',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
    install_requires=[
        'django>=3.1,<4.0',
        'django-ipware>=3.0.2',
        'ansi2html>=1.6.0',
        'django-chamber>=0.6.16',
        'django-choice-enumfields>=1.1.2',
        'django-generic-m2m-field>=0.0.9',
        'django-celery-extensions>=0.0.37',
        'isodate>=0.6.1',
    ],
    zip_safe=False
)



from security.backends.app import SecurityBackend


class SecurityTestingBackend(SecurityBackend):

    name = 'security.backends.testing'
    label = 'security_backends_testing'
    backend_name = 'testing'
    reader = 'security.backends.testing.reader.TestingBackendReader'



from security.utils import update_logged_request_data


class LogDjangoModelResourceMixin:

    log_update_obj = log_get_obj = True

    def _post_save_obj(self, obj, form, change):
        super()._post_save_obj(obj, form, change)
        if self.log_update_obj:
            update_logged_request_data(self.request, related_objects=[obj])

    def _get_obj_or_404(self, pk=None):
        obj = super()._get_obj_or_404(pk)
        if self.log_get_obj:
            update_logged_request_data(self.request, related_objects=[obj])
        return obj



try:
    import json
except ImportError:
    import simplejson as json


def serialize_message(message, metadata):
    return bytes(
        json.dumps({
            'message': message,
            'meta': metadata,
        }) + '\n', 'utf-8'
    )



from security.backends.app import SecurityBackend


class SecurityLoggingBackend(SecurityBackend):

    name = 'security.backends.logging'
    label = 'security_backends_logging'
    backend_name = 'logging'
    writer = 'security.backends.logging.writer.LoggingBackendWriter'



from urllib.parse import urlparse

from django.template.defaultfilters import truncatechars
from django.urls.exceptions import Resolver404
from django.urls import resolve
from django.utils.timezone import now

from ipware.ip import get_client_ip

from security.enums import LoggerName
from security.config import settings
from security.backends.signals import (
    input_request_started, input_request_finished, input_request_error,
    output_request_started, output_request_finished, output_request_error,
)
from security.logging.common import SecurityLogger

from .utils import clean_headers, clean_queries, clean_body


class InputRequestLogger(SecurityLogger):

    logger_name = LoggerName.INPUT_REQUEST

    def __init__(self, hide_request_body=False, request_headers=None, request_body=None, user_id=None, method=None,
                 host=None, path=None, queries=None, is_secure=None, ip=None, view_slug=None, response_body=None,
                 response_code=None, response_headers=None, **kwargs):
        super().__init__(**kwargs)
        self._hide_request_body = hide_request_body
        self.request_headers = request_headers
        self.request_body = request_body
        self.user_id = user_id
        self.method = method
        self.host = host
        self.path = path
        self.queries = queries
        self.is_secure = is_secure
        self.ip = ip
        self.view_slug = view_slug
        self.response_body = response_body
        self.response_code = response_code
        self.response_headers = response_headers

    def log_request(self, request):
        try:
            view_slug = resolve(request.path_info, getattr(request, 'urlconf', None)).view_name
        except Resolver404:
            view_slug = None

        self.request_headers = clean_headers(request.headers)
        self.request_body = (
            settings.SENSITIVE_DATA_REPLACEMENT
            if request.body and self._hide_request_body
            else clean_body(request.body, settings.LOG_REQUEST_BODY_LENGTH)
        )
        self.user_id = request.user.pk if hasattr(request, 'user') and request.user.is_authenticated else None
        self.method = request.method.upper()[:7]
        self.host = request.get_host()
        self.path = truncatechars(request.path, 200)
        self.queries = clean_queries(request.GET.dict())
        self.is_secure = request.is_secure()
        self.ip = get_client_ip(request)[0]
        self.start = now()
        self.view_slug = view_slug
        input_request_started.send(sender=InputRequestLogger, logger=self)

    def log_response(self, request, response):
        if (not response.streaming
                and (settings.LOG_RESPONSE_BODY_CONTENT_TYPES is None
                     or response.get('content-type', '').split(';')[0] in settings.LOG_RESPONSE_BODY_CONTENT_TYPES)):
            self.response_body = clean_body(response.content, settings.LOG_RESPONSE_BODY_LENGTH)
        else:
            self.response_body = None

        self.stop = now()
        self.response_code = response.status_code
        self.response_headers = clean_headers(dict(response.items()))
        self.user_id = request.user.pk if hasattr(request, 'user') and request.user.is_authenticated else None
        input_request_finished.send(sender=InputRequestLogger, logger=self)

    def log_exception(self, ex_tb):
        self.error_message = ex_tb
        input_request_error.send(sender=InputRequestLogger, logger=self)


class OutputRequestLogger(SecurityLogger):

    logger_name = LoggerName.OUTPUT_REQUEST

    def __init__(self, request_headers=None, request_body=None, method=None, host=None, path=None, queries=None,
                 is_secure=None, response_body=None, response_code=None, response_headers=None, **kwargs):
        super().__init__(**kwargs)
        self.request_headers = request_headers
        self.request_body = request_body
        self.method = method
        self.host = host
        self.path = path
        self.queries = queries
        self.is_secure = is_secure
        self.response_body = response_body
        self.response_code = response_code
        self.response_headers = response_headers

    def log_request(self, request):
        from .utils import get_logged_params

        parsed_url = urlparse(request.url)
        self.is_secure = parsed_url.scheme == 'https'
        self.host = parsed_url.netloc
        self.path = parsed_url.path
        self.method = request.method.upper()[:7]
        self.queries = clean_queries(get_logged_params(request.url))
        self.start = now()
        self.request_headers = clean_headers(dict(request.headers.items()))
        self.request_body = clean_body(request.body, settings.LOG_REQUEST_BODY_LENGTH)
        output_request_started.send(sender=OutputRequestLogger, logger=self)

    def log_response(self, response):
        self.stop = now()
        self.response_code = response.status_code
        self.response_headers = clean_headers(dict(response.headers.items()))
        self.response_body = clean_body(response.content, settings.LOG_RESPONSE_BODY_LENGTH)
        output_request_finished.send(sender=OutputRequestLogger, logger=self)

    def log_exception(self, ex_tb):
        self.error_message = ex_tb
        self.stop = now()
        output_request_error.send(sender=InputRequestLogger, logger=self)



from datetime import timedelta

from django.core.management.base import CommandError
from django.core.management.base import BaseCommand
from django.utils.timezone import now


UNIT_OPTIONS = {
    'h': lambda amount: now() - timedelta(hours=amount),
    'd': lambda amount: now() - timedelta(days=amount),
    'w': lambda amount: now() - timedelta(weeks=amount),
    'm': lambda amount: now() - timedelta(days=(30 * amount)),  # 30-day month
    'y': lambda amount: now() - timedelta(weeks=(52 * amount)),  # 364-day year
}


class Command(BaseCommand):

    models = {}

    def add_arguments(self, parser):
        parser.add_argument('--type', action='store', dest='type', default='input-request',
                            help='Tells Django what type of requests should be removed '
                                 '({}).'.format('/'.join(self.models.keys())))
        parser.add_argument('--expiration', action='store', dest='expiration',
                            help='Sets the timedelta from which logs will be removed.', required=True)
        parser.add_argument('--noinput', action='store_false', dest='interactive', default=True,
                            help='Tells Django to NOT prompt the user for input of any kind.')

    def _clean_data(self, qs, options):
        raise NotImplementedError

    def _get_queryset(self, model, timestamp):
        raise NotImplementedError

    def _get_qs_count(self, qs):
        raise NotImplementedError

    def handle(self, expiration, type, **options):
        # Check we have the correct values
        unit = expiration[-1]
        amount = expiration[0:-1]
        try:
            amount = int(amount)
        except ValueError:
            raise CommandError('Invalid expiration format')

        if unit not in UNIT_OPTIONS:
            raise CommandError('Invalid expiration format')

        model = self._get_model(type)
        qs = self._get_queryset(model, UNIT_OPTIONS[unit](amount))
        if self._get_qs_count(qs) == 0:
            self.stdout.write('There are no logs to delete.')
        else:
            if options.get('interactive'):
                confirm = input('''
                   You have requested a database reset.
                   This will IRREVERSIBLY DESTROY any
                   logs created before {} {}
                   ago. That is a total of {} logs.
                   Are you sure you want to do this?
                   Type 'yes' to continue, or 'no' to cancel: '''.format(amount, unit, qs.count()))
            else:
                confirm = 'yes'

            if confirm == 'yes':
                self.stdout.write('Clean data')
                self._clean_data(qs, options)

    def _get_model(self, type):
        model = self.models.get(type)
        if model:
            return model
        else:
            raise CommandError('Type can be only one of "{}"'.format('/'.join(self.models.keys())))



import re
from importlib import import_module
from io import StringIO
from time import time
import datetime
import decimal
import uuid

import isodate

from django.core.exceptions import ImproperlyConfigured
from django.db import router
from django.utils.timezone import now, is_aware
from django.utils.duration import duration_iso_string

from .config import settings
from .enums import LoggerName


def get_throttling_validators(name):
    try:
        return getattr(import_module(settings.DEFAULT_THROTTLING_VALIDATORS_PATH), name)
    except (ImportError, AttributeError):
        raise ImproperlyConfigured('Throttling validator configuration {} is not defined'.format(name))


def remove_nul_from_string(value):
    return value.replace('\x00', '')


def is_atty_string(s):
    return bool(re.search(r'^' + re.escape('\x1b[') + r'\d+m$', s))


def truncate_lines_from_left(value, max_length):
    """
    Firstly text is truncated according to max_length.
    Next step is find next character '\n' and while text before \n is removed. On the start is added '…'

    If there is only one line (without \n) which is longer than max_length. Whole line is removed and replaced with …\n
    """
    if len(value) <= max_length:
        return value

    # value 2 is added because of …\n
    truncated_value = value[len(value) + 2 - max_length:]
    truncated_value_split_by_newline = truncated_value.split('\n', 1)
    return f'…\n{"" if len(truncated_value_split_by_newline) == 1 else truncated_value_split_by_newline[1]}'


class LogStringIO(StringIO):

    BOLD = '\x1b[1m'
    RESET = '\x1b[0m'

    def __init__(self, flush_callback=None, flush_timeout=None, write_time=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_closed = False
        self._last_newline = 0
        self._flush_callback = flush_callback
        self._write_time = write_time
        self._write_time_to_line = self._write_time
        self._last_flush_time = time()
        self._flush_timeout = settings.LOG_STRING_IO_FLUSH_TIMEOUT if flush_timeout is None else flush_timeout
        self._flushed = True
        self._next_line_number = 1

    def _flush(self, force=False):
        if (not self._flushed and self._flush_callback
                and (force or time() - self._last_flush_time > self._flush_timeout)):
            self._flush_callback(self)
            self._last_flush_time = time()
            self._flushed = True

    def _write_line_number_and_time(self):
        if self._write_time:
            super().write('{}{}[{}] [{}]{} '.format(
                self.RESET,
                self.BOLD,
                self._next_line_number,
                now().strftime('%d-%m-%Y %H:%M:%S'),
                self.RESET
            ))
            self._write_time_to_line = False

    def write(self, s):
        if not self._is_closed:
            self._flushed = False
            lines = s.split('\n')

            for i, line in enumerate(lines):
                cursor_first = True
                for cursor_block in line.split('\r'):
                    if not cursor_first:
                        self.seek(self._last_newline)
                        self.truncate()
                        self._write_time_to_line = self._write_time
                    cursor_first = False
                    if self._write_time_to_line and not is_atty_string(cursor_block):
                        self._write_line_number_and_time()
                    super().write(cursor_block)

                if i != len(lines) - 1:
                    super().write('\n')
                    self._next_line_number += 1
                    self._last_newline = self.tell()
                    self._write_time_to_line = self._write_time

            self._truncate_by_max_length()
            self._flush()

    def _truncate_by_max_length(self):
        """
        StringIO value is truncated from left side to maximal length defined in LOG_STRING_OUTPUT_TRUNCATE_LENGTH.
        Because too frequent string truncation can cause high CPU load, log string is truncated by more characters.
        The value is defined in LOG_STRING_OUTPUT_TRUNCATE_OFFSET. Eg
            LOG_STRING_OUTPUT_TRUNCATE_LENGTH = 10000
            LOG_STRING_OUTPUT_TRUNCATE_OFFSET = 1000

            if text is longer than 10000 character is truncated to 9000 characters.

        To better readability the text is further truncated to the following newline.
        """
        if self.tell() > settings.LOG_STRING_OUTPUT_TRUNCATE_LENGTH:
            original_value = self.getvalue()
            truncated_value = truncate_lines_from_left(
                self.getvalue(),
                settings.LOG_STRING_OUTPUT_TRUNCATE_LENGTH - settings.LOG_STRING_OUTPUT_TRUNCATE_OFFSET
            )
            self._last_newline = self._last_newline - (len(original_value) - len(truncated_value))
            self.seek(0)
            self.truncate()
            super().write(truncated_value)
            if truncated_value.endswith('\n'):
                self._write_line_number_and_time()

    def isatty(self):
        return True

    def close(self):
        self._is_closed = True
        self._flush(force=True)
        super().close()


class TeeStringIO:
    """
    StringIO that additionally writes to another stream(s).
    """
    ANSI_ESCAPE_REGEX = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def __init__(self, out, log_stream):
        self._out = out
        self._log_stream = log_stream
        self._last_newline = 0
        super().__init__()

    def write(self, s):
        if self.isatty():
            self._out.write(s)
        else:
            self._out.write(self.ANSI_ESCAPE_REGEX.sub('', s))
        self._log_stream.write(s)

    def __getattr__(self, name):
        return getattr(self._out, name)

    def isatty(self):
        return hasattr(self._out, 'isatty') and self._out.isatty()

    def flush(self):
        self._out.flush()
        self._log_stream.flush()


def regex_sub_groups_global(pattern, repl, string):
    """
    Globally replace all groups inside pattern with `repl`.
    If `pattern` doesn't have groups the whole match is replaced.
    """
    for search in reversed(list(re.finditer(pattern, string))):
        for i in range(len(search.groups()), 0 if search.groups() else -1, -1):
            start, end = search.span(i)
            string = string[:start] + repl + string[end:]
    return string


def update_logged_request_data(request, related_objects=None, slug=None, extra_data=None):
    input_request_logger = getattr(request, 'input_request_logger', None)
    if input_request_logger:
        if related_objects:
            input_request_logger.add_related_objects(*related_objects)
        if slug:
            input_request_logger.set_slug(slug)
        if extra_data:
            input_request_logger.update_extra_data(extra_data)


def get_object_triple(obj):
    from django.contrib.contenttypes.models import ContentType

    if isinstance(obj, (list, tuple)) and len(obj) == 3:
        return tuple(obj)
    else:
        content_type = ContentType.objects.get_for_model(obj)
        return router.db_for_write(content_type.model_class()), content_type.pk, obj.pk


def serialize_data(o):
    if isinstance(o, datetime.datetime):
        r = o.isoformat()
        if o.microsecond:
            r = r[:23] + r[26:]
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return {'@type': 'datetime',  '@value': r}
    elif isinstance(o, datetime.date):
        return {'@type': 'date',  '@value': o.isoformat()}
    elif isinstance(o, datetime.time):
        if is_aware(o):
            raise ValueError("JSON can't represent timezone-aware times.")
        r = o.isoformat()
        if o.microsecond:
            r = r[:12]
        return {'@type': 'time', '@value': r}
    elif isinstance(o, datetime.timedelta):
        return {'@type': 'timedelta', '@value': duration_iso_string(o)}
    elif isinstance(o, decimal.Decimal):
        return {'@type': 'decimal', '@value': str(o)}
    elif isinstance(o, uuid.UUID):
        return {'@type': 'uuid', '@value': str(o)}
    elif isinstance(o, (list, tuple, set)):
        return [serialize_data(v) for v in o]
    elif isinstance(o, dict):
        return {
            k: serialize_data(v) for k, v in o.items()
        }
    else:
        return o


def deserialize_data(o):
    if isinstance(o, dict) and set(o.keys()) == {'@type', '@value'}:
        o_type, o_value = o['@type'], o['@value']
        if o_type == 'uuid':
            return uuid.UUID(o_value)
        elif o_type == 'decimal':
            return decimal.Decimal(o_value)
        elif o_type == 'timedelta':
            return isodate.parse_duration(o_value)
        elif o_type == 'time':
            return isodate.parse_time(o_value)
        elif o_type == 'date':
            return isodate.parse_date(o_value)
        elif o_type == 'datetime':
            return isodate.parse_datetime(o_value)
    elif isinstance(o, dict):
        return {
            k: deserialize_data(v) for k, v in o.items()
        }
    elif isinstance(o, list):
        return [deserialize_data(v) for v in o]
    else:
        return o


def get_run_logger():
    """
    Return celery task run or command logger
    """
    from security.logging.common import get_last_logger

    return get_last_logger(LoggerName.CELERY_TASK_RUN) or get_last_logger(LoggerName.COMMAND)


def add_related_objects_to_run_logger(obj):
    """
    Add related object to celery task run or command logger
    """
    run_logger = get_run_logger()
    if run_logger:
        run_logger.add_related_objects(obj)
